#pragma once
#include <Arduino.h>
#include <list>
// #define IRDEBUG_INFO
/*//////////////////////////////////////////////////////////////////////////////////////

Для работы в паре положить декодер в энкодер

*/
// Адресация с 1 до 65 499
#define IR_Broadcast 65000 // 65 500 ~ 65 535 - широковещательные пакеты (всем)
/*
*Адресное пространство:
    Адрес 0 запрещен и зарезервирован под NULL, либо тесты
    IR_MSG_ACCEPT с адреса 0 воспринимается всеми устройствами
*/ 
//**** Контрольные точки ******
#define IR_MAX_ADDR_CPU 64999
#define IR_MIN_ADDR_CPU 32000

// //***** Группы машинок ********
// #define IR_MAX_CAR_GROUP 31999
// #define IR_MIN_CAR_GROUP 30000

// //********** FREE *************
// #define IR_MAX_FREE 31999
// #define IR_MIN_FREE 2000

//********* Машинки ***********
#define IR_MAX_CAR 31999
#define IR_MIN_CAR 100

//***** Пульты управления *****
#define IR_MAX_CONTROLLER 99
#define IR_MIN_CONTROLLER 0
/*

/```````````````````````````````````````````````` data pack `````````````````````````````````````````````\                                  
                                                                                                         
{``````````} [````````````````````````] [````````````````````] [````````````````````````] [``````````````]
{ msg type } [   addr_from uint16_t   ] [  addr_to uint16_t  ] [====== data bytes ======] [   CRC Bytes  ]
{..........} [........................] [....................] [........................] [..............]
                                                                                                          
{ aka size } [addr_from_H][addr_from_L] [addr_to_H][addr_to_L] [data_H][data_n..][data_L] [ crc1 ][ crc2 ]
|     0           1            2              3         4          5                         |       |    
\____________________________________________________________________________________________/       |    
|                                                                                                    |    
\____________________________________________________________________________________________________/    

msg type:
                                        //   __________
                                        //  | 01234567 |
                                        //   ----------
                                        //  | xxx..... | = тип сообщения
                                        //  | ...xxxxx | = длина (максимум 31 бита)
                                        //   ---------- */
#define IR_MSG_BACK 0U          //  | 000...... | = Задний сигнал машинки
#define IR_MSG_ACCEPT 1U        //  | 001..... | = подтверждение
#define IR_MSG_REQUEST 2U       //  | 010..... | = запрос
#define IR_MSG_ 3U              //  | 011..... | = ??
#define IR_MSG_BACK_TO 4U       //  | 100..... | = Задний сигнал машинки c адресацией
#define IR_MSG_ 5U              //  | 101..... | = ??
#define IR_MSG_DATA_NOACCEPT 6U //  | 110..... | = данные, не требующие подтверждения
#define IR_MSG_DATA_ACCEPT 7U   //  | 111..... | = данные требующие подтверждения
;                               /*                                     //   ----------
                              
                               /``````````````````````````````` подтверждение `````````````````````````````\      /``````````````````````````````````````` запрос ``````````````````````````````````\
                                                                                                                                                     
                               {``````````} [````````````````````````] [``````````````````] [``````````````]      {``````````} [````````````````````````] [````````````````````````] [``````````````]
                               { msg type } [   addr_from uint16_t   ] [=== customByte ===] [   CRC Bytes  ]      { msg type } [   addr_from uint16_t   ] [    addr_to uint16_t    ] [   CRC Bytes  ]
                               {..........} [........................] [..................] [..............]      {..........} [........................] [........................] [..............]
                                                                                                                                                                                           
                               { 001..... } [addr_from_H][addr_from_L] [=== customByte ===] [ crc1 ][ crc2 ]      { 010..... } [addr_from_H][addr_from_L] [addr_from_H][addr_from_L] [ crc1 ][ crc2 ]
                               |     0            1           2                  3              4       5          |     0            1           2              3           4           5       6    
                               \________________________________________________________________/       |          \_____________________________________________________________________/       |    
                               |                                                                        |          |                                                                             |    
                               \________________________________________________________________________/          \_____________________________________________________________________________/    
                              
                               customByte - контрольная сумма принятых данных по poly1
                              
                              
                              
                               /`````````````````````` Задний сигнал машинки без адресации ``````````````````````\        
                                                                                                                          
                               {``````````} [````````````````````````] [````````````````````````] [``````````````]        
                               { msg type } [   addr_from uint16_t   ] [====== data bytes ======] [   CRC Bytes  ]        
                               {..........} [........................] [........................] [..............]        
                                                                                                                          
                               { 0000xxxx } [addr_from_H][addr_from_L] [data_H][data_n..][data_L] [ crc1 ][ crc2 ]        
                               |     0           1            2            3                         |       |            
                               \_____________________________________________________________________/       |            
                               |                                                                             |            
                               \_____________________________________________________________________________/            
                              
                              
                              
                               /```````````````````````````````````` Задний сигнал машинки с адресацией ````````````````````````````````````\ 
                                                                                                                   
                               {``````````} [````````````````````````] [````````````````````````] [````````````````````````] [``````````````] 
                               { msg type } [   addr_from uint16_t   ] [    addr_to uint16_t    ] [====== data bytes ======] [   CRC Bytes  ] 
                               {..........} [........................] [........................] [........................] [..............] 
                                                                                                                                              
                               { 0001xxxx } [addr_from_H][addr_from_L] [addr_from_H][addr_from_L] [data_H][data_n..][data_L] [ crc1 ][ crc2 ] 
                               |     0           1            2              3           4            5                         |       |     
                               \________________________________________________________________________________________________/       |     
                               |                                                                                                        |     
                               \________________________________________________________________________________________________________/     
                              
                               */

#define IR_MASK_MSG_TYPE 0b00000111
#define IR_MASK_MSG_INFO 0b00011111

/*
/////////////////////////////////////////////////////////////////////////////////////*/
typedef uint16_t crc_t;

// #define BRUTEFORCE_CHECK                            // Перепроверяет пакет на 1 битные ошибки //TODO: зависает
#define bytePerPack 16 // колличество байтов в пакете
#ifndef freeFrec
#define freeFrec false
#endif

#ifndef subBufferSize
#define subBufferSize 50 // Буфер для складирования фронтов, пока их не обработают (передатчик)
#endif

#define preambPulse 3

#define disablePairDec false // Отключать парный приёмник, возможны баги, используйте setBlindDecoders()

/////////////////////////////////////////////////////////////////////////////////////

#define bitPerByte 8U // Колличество бит в байте
#define addrBytes 2
#define msgBytes 1
#define crcBytes 2
#define poly1 0x31
#define poly2 0x8C
#define syncBits 3U // количество битов синхронизации

#define dataByteSizeMax (msgBytes + addrBytes + addrBytes + bytePerPack + crcBytes)

#define preambFronts (preambPulse * 2)                              // количество фронтов преамбулы (Приём)
#define preambToggle ((bitPauseTakts * 2 + bitActiveTakts) * 2 - 1) // колличество переключений преамбулы (Передача)

#define carrierFrec 38000U                     // частота несущей (Приём/Передача)
#define carrierPeriod (1000000U / carrierFrec) // период несущей в us (Приём)

// В процессе работы значения будут отклонятся в соответствии с предыдущим битом
#define bitActiveTakts 25U // длительность высокого уровня в тактах
#define bitPauseTakts 12U  // длительность низкого уровня в тактах

#define bitTakts (bitActiveTakts + bitPauseTakts) // Общая длительность бита в тактах
#define bitTime (bitTakts * carrierPeriod)        // Общая длительность бита
#define tolerance 300U

constexpr uint16_t test_all_Time = bitTime;
constexpr uint16_t test_all_Takts = bitTakts * 2;
constexpr uint16_t test_hi = ((bitPauseTakts) * 2 - 0) + ((bitActiveTakts) * 2 - 0);
constexpr uint16_t test_low = ((bitPauseTakts / 2 + bitActiveTakts) * 2 - 0) + ((bitPauseTakts)-0);

class IR_FOX
{
public:
    struct PackOffsets
    {
        uint8_t msgOffset;
        uint8_t addrFromOffset;
        uint8_t addrToOffset;
        uint8_t dataOffset;
        uint8_t crcOffset;
    };

    struct ErrorsStruct
    {
        uint8_t lowSignal = 0;
        uint8_t highSignal = 0;
        uint8_t other = 0;

        void reset()
        {
            lowSignal = 0;
            highSignal = 0;
            other = 0;
        }
        uint16_t all() { return lowSignal + highSignal + other; }
    };

    struct PackInfo
    {
        uint8_t *buffer = nullptr;
        uint8_t packSize = 0;
        uint16_t crc = 0;
        ErrorsStruct err;
        uint16_t rTime = 0;
    };

    inline uint16_t getId() const { return id; }
    inline void setId(uint16_t id) { this->id = id; }
    static void checkAddressRuleApply(uint16_t address, uint16_t id, bool &flag);
    void setPin(uint8_t pin);
    inline uint8_t getPin() { return pin; };

protected:
    uint16_t id;
    uint8_t pin;
    GPIO_TypeDef *port;
    uint16_t mask;
    ErrorsStruct errors;
    uint8_t crc8(uint8_t *data, uint8_t start, uint8_t end, uint8_t poly);
};
